/*
** $Id: lua.c,v 1.55 2000/10/20 16:36:32 roberto Exp $
** Lua stand-alone interpreter
** See Copyright Notice in lua.h
*/

/** Portions backported from lua.c v1.71:
 **
 ** readline
 ** incomplete line handling -- well, got stalled out on that, not done.
 ** = for print shortcut
 **/

/* backport goo */

#define l_s(x) x
#define l_c(x) x
#define l_char char


#define DLUA_INIT_FILE "/etc/dlua4.0/init.lua"

#define DLUA_VERSION "DLua 4.0 - Debian packaged Lua"
/* really no point in splatting another banner here. */
#define DLUA_COPYRIGHT ""

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lua.h"

#include "luadebug.h"
#include "lualib.h"

#include "lua_loadlib2.h"

static lua_State *L = NULL;


#ifndef PROMPT
#define PROMPT		"> "
#endif

#ifndef PROMPT2
#define PROMPT2		l_s(">> ")
#endif


#ifdef _POSIX_SOURCE
#include <unistd.h>
#else
static int isatty (int x) { return x==0; }  /* assume stdin is a tty */
#endif


/*
** global options
*/
struct Options {
  int toclose;
  int stacksize;
};


typedef void (*handler)(int);  /* type for signal actions */

static void laction (int i);


static lua_Hook old_linehook = NULL;
static lua_Hook old_callhook = NULL;


static void userinit (void) {
  lua_baselibopen(L);
  lua_iolibopen(L);
  lua_strlibopen(L);
  lua_mathlibopen(L);
  lua_dblibopen(L);
  /* add your libraries here */
  lua_loadlib2_open(L);
}

static void dlua_init(void) 
{
  int oldtop = lua_gettop(L);
  int result = lua_dofile(L, DLUA_INIT_FILE);
  if (result) {
    fprintf(stderr, "could not execute dlua init file " DLUA_INIT_FILE "\n");
    exit(1);
  }
  lua_settop(L, oldtop);
}

static handler lreset (void) {
  return signal(SIGINT, laction);
}


static void lstop (void) {
  lua_setlinehook(L, old_linehook);
  lua_setcallhook(L, old_callhook);
  lreset();
  lua_error(L, "interrupted!");
}


static void laction (int i) {
  (void)i;  /* to avoid warnings */
  signal(SIGINT, SIG_DFL); /* if another SIGINT happens before lstop,
                              terminate process (default action) */
  old_linehook = lua_setlinehook(L, (lua_Hook)lstop);
  old_callhook = lua_setcallhook(L, (lua_Hook)lstop);
}


static int ldo (int (*f)(lua_State *l, const char *), const char *name, 
                int clear) {
  int res;
  handler h = lreset();
  int top = lua_gettop(L);
  res = f(L, name);  /* dostring | dofile */
  if (clear) lua_settop(L, top);  /* remove eventual results */
  signal(SIGINT, h);  /* restore old action */
  /* Lua gives no message in such cases, so lua.c provides one */
  if (res == LUA_ERRMEM) {
    fprintf(stderr, "lua: memory allocation error\n");
  }
  else if (res == LUA_ERRERR)
    fprintf(stderr, "lua: error in error message\n");
  return res;
}


static void print_message (void) {
  fprintf(stderr,
  "usage: lua [options].  Available options are:\n"
  "  -        execute stdin as a file\n"
  "  -c       close Lua when exiting\n"
  "  -e stat  execute string `stat'\n"
  "  -f name  execute file `name' with remaining arguments in table `arg'\n"
  "  -i       enter interactive mode with prompt\n"
  "  -q       enter interactive mode without prompt\n"
  "  -sNUM    set stack size to NUM (must be the first option)\n"
  "  -v       print version information\n"
  "  a=b      set global `a' to string `b'\n"
  "  name     execute file `name'\n"
);
}


static void print_version (void) {
  printf("%.80s  %.80s\n", LUA_VERSION, LUA_COPYRIGHT);
  printf("%.80s  %.80s\n", DLUA_VERSION, DLUA_COPYRIGHT);
}


static void assign (char *arg) {
  char *eq = strchr(arg, '=');
  *eq = '\0';  /* spilt `arg' in two strings (name & value) */
  lua_pushstring(L, eq+1);
  lua_setglobal(L, arg);
}


static void getargs (char *argv[]) {
  int i;
  lua_newtable(L);
  for (i=0; argv[i]; i++) {
    /* arg[i] = argv[i] */
    lua_pushnumber(L, i);
    lua_pushstring(L, argv[i]);
    lua_settable(L, -3);
  }
  /* arg.n = maximum index in table `arg' */
  lua_pushstring(L, "n");
  lua_pushnumber(L, i-1);
  lua_settable(L, -3);
}


static int l_getargs (lua_State *l) {
  char **argv = (char **)lua_touserdata(l, -1);
  getargs(argv);
  return 1;
}


static int file_input (const char *argv) {
  int result = ldo(lua_dofile, argv, 1);
  if (result) {
    if (result == LUA_ERRFILE) {
      fprintf(stderr, "lua: cannot execute file ");
      perror(argv);
    }
    return EXIT_FAILURE;
  }
  else
    return EXIT_SUCCESS;
}


#ifdef USE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#define save_line(b)	if (b[0]) add_history(b)
#define free_line(b)	free(b)
#else
#define save_line(b)
#define free_line(b)

/* maximum length of an input line */
#ifndef MAXINPUT
#define MAXINPUT	BUFSIZ
#endif


static l_char *readline (const l_char *prompt) {
  static l_char buffer[MAXINPUT];
  if (prompt) {
    fputs(prompt, stdout);
    fflush(stdout);
  }
  return fgets(buffer, sizeof(buffer), stdin);
}
#endif

static const l_char *get_prompt (const l_char *name, const l_char *def) {
  const l_char *p = NULL;
  lua_getglobal(L, name);
  p = lua_tostring(L, -1);
  if (p == NULL) p = def;
  lua_pop(L, 1);  /* remove global */
  return p;
}

static l_char *get_line (int prompt, int incomplete) {
  const l_char *p = NULL;
  if (prompt) {
    if (incomplete)
      p = get_prompt(l_s("_PROMPT2"), PROMPT2);
    else
      p = get_prompt(l_s("_PROMPT"), PROMPT);
  }
  return readline(p);
}

static int incomplete = 0;

#if 1
int is_incomplete(lua_State *l, char *s, int len) {
  return 0;
}
#else
extern int is_incomplete(lua_State *l, char *s, int len);
#endif

static int load_string (int prompt, int *toprint) {
#if 0
  lua_getglobal(L, l_s(LUA_ERRORMESSAGE));
  lua_register(L, l_s(LUA_ERRORMESSAGE), trap_eof);
#endif
  incomplete = 0;
  lua_pushstring(L, "");
  for (;;) {
    /* int result; */
    l_char *buffer = get_line(prompt,incomplete);
    if (buffer == NULL) {
      lua_settop(L, 1);
      /* lua_setglobal(L, l_s(LUA_ERRORMESSAGE)); */
      return 0;
    }
    { 
      int len = strlen(buffer);
      if (len > 1 && buffer[len-2] == '\\') {
        incomplete = 1;
        buffer[len-2] = '\0';
      } else {
        incomplete = 0;
      }
    }
    save_line(buffer);
    if (!incomplete && buffer[0] == l_c('=')) {
      buffer[0] = l_c(' ');
      lua_pushstring(L, l_s("return"));
      *toprint = 1;
    }
    lua_pushstring(L, buffer);
    free_line(buffer);
    lua_concat(L, lua_gettop(L));
    /* incomplete = 0; */
    /* result = lua_loadbuffer(L, lua_tostring(L, 2), lua_strlen(L, 2), "=stdin"); */
#if 0
    incomplete = is_incomplete(L, lua_tostring(L,-1), lua_strlen(L, -1));
#endif
    if (incomplete) continue;
#if 0
    lua_remove(L, 2);
    if (result == 0) {
      lua_insert(L, 1);  /* swap compiled chunk with old _ERRORMESSAGE */
      lua_setglobal(L, l_s(LUA_ERRORMESSAGE));
      return 1;
    }
    else
      report(result);
#endif
    return 1;
 }
}


#if 0
static int lcall (lua_State *l, const l_char *name) {
  (void)name;  /* to avoid warnings */
  return lua_call(l, 0, LUA_MULTRET);
}
#endif

static void manual_input (int version, int prompt) {
  int toprint = 0;
  if (version) print_version();
  while (load_string(prompt, &toprint)) {
    ldo(lua_dostring, lua_tostring(L, -1), 0);
    lua_remove(L, 1);
    if (toprint && lua_gettop(L) > 0) {  /* any result to print? */
      lua_getglobal(L, l_s("print"));
      lua_insert(L, 1);
      lua_call(L, lua_gettop(L)-1, 0);
      toprint = 0;
    }
    else
      lua_settop(L, 0);  /* remove eventual results */
  }
  printf(l_s("\n"));
}

static int handle_argv (char *argv[], struct Options *opt) {
  if (opt->stacksize > 0) argv++;  /* skip option `-s' (if present) */
  if (*argv == NULL) {  /* no more arguments? */
    if (isatty(0)) {
      manual_input(1, 1);
    }
    else
      ldo(lua_dofile, NULL, 1);  /* executes stdin as a file */
  }
  else {  /* other arguments; loop over them */
    int i;
    for (i = 0; argv[i] != NULL; i++) {
      if (argv[i][0] != '-') {  /* not an option? */
        if (strchr(argv[i], '='))
          assign(argv[i]);
        else
          if (file_input(argv[i]) != EXIT_SUCCESS)
            return EXIT_FAILURE;  /* stop if file fails */
        }
        else switch (argv[i][1]) {  /* option */
          case 0: {
            ldo(lua_dofile, NULL, 1);  /* executes stdin as a file */
            break;
          }
          case 'i': {
            manual_input(0, 1);
            break;
          }
          case 'q': {
            manual_input(0, 0);
            break;
          }
          case 'c': {
            opt->toclose = 1;
            break;
          }
          case 'v': {
            print_version();
            break;
          }
          case 'e': {
            i++;
            if (argv[i] == NULL) {
              print_message();
              return EXIT_FAILURE;
            }
            if (ldo(lua_dostring, argv[i], 1) != 0) {
              fprintf(stderr, "lua: error running argument `%.99s'\n", argv[i]);
              return EXIT_FAILURE;
            }
            break;
          }
          case 'f': {
            i++;
            if (argv[i] == NULL) {
              print_message();
              return EXIT_FAILURE;
            }
            getargs(argv+i);  /* collect remaining arguments */
            lua_setglobal(L, "arg");
            return file_input(argv[i]);  /* stop scanning arguments */
          }
          case 's': {
            fprintf(stderr, "lua: stack size (`-s') must be the first option\n");
            return EXIT_FAILURE;
          }
          default: {
            print_message();
            return EXIT_FAILURE;
          }
        }
    }
  }
  return EXIT_SUCCESS;
}


static void getstacksize (int argc, char *argv[], struct Options *opt) {
  if (argc >= 2 && argv[1][0] == '-' && argv[1][1] == 's') {
    int stacksize = atoi(&argv[1][2]);
    if (stacksize <= 0) {
      fprintf(stderr, "lua: invalid stack size ('%.20s')\n", &argv[1][2]);
      exit(EXIT_FAILURE);
    }
    opt->stacksize = stacksize;
  }
  else
    opt->stacksize = 0;  /* no stack size */
}


static void register_getargs (char *argv[]) {
  lua_pushuserdata(L, argv);
  lua_pushcclosure(L, l_getargs, 1);
  lua_setglobal(L, "getargs");
}


int main (int argc, char *argv[]) {
  struct Options opt;
  int status;
  opt.toclose = 0;
  getstacksize(argc, argv, &opt);  /* handle option `-s' */
  L = lua_open(opt.stacksize);  /* create state */
  userinit();  /* open libraries */
  dlua_init();
  register_getargs(argv);  /* create `getargs' function */
  status = handle_argv(argv+1, &opt);
  if (opt.toclose)
    lua_close(L);
  return status;
}

